import contextlib

import pytest

from aiojenkins.exceptions import JenkinsError, JenkinsNotFoundError
from tests import jenkins


@pytest.mark.asyncio
async def test_get_nodes():
    await jenkins.nodes.get_all()


@pytest.mark.asyncio
async def test_get_node_info():
    info = await jenkins.nodes.get_info('(master)')
    assert isinstance(info, dict)
    info = await jenkins.nodes.get_info('master')
    assert isinstance(info, dict)


@pytest.mark.asyncio
async def test_is_node_exists():
    is_exists = await jenkins.nodes.is_exists('master')
    assert is_exists is True

    is_exists = await jenkins.nodes.is_exists('random_empty_node')
    assert is_exists is False

    is_exists = await jenkins.nodes.is_exists('')
    assert is_exists is False


@pytest.mark.asyncio
async def test_disable_node():
    for _ in range(2):
        await jenkins.nodes.disable('master')
        info = await jenkins.nodes.get_info('master')
        assert info['offline'] is True


@pytest.mark.asyncio
async def test_enable_node():
    for _ in range(2):
        await jenkins.nodes.enable('master')
        info = await jenkins.nodes.get_info('master')
        assert info['offline'] is False


@pytest.mark.asyncio
async def test_update_node_offline_reason():
    await jenkins.nodes.update_offline_reason('master', 'maintenance1')
    info = await jenkins.nodes.get_info('master')
    assert info['offlineCauseReason'] == 'maintenance1'

    await jenkins.nodes.update_offline_reason('master', 'maintenance2')
    info = await jenkins.nodes.get_info('master')
    assert info['offlineCauseReason'] == 'maintenance2'


@pytest.mark.asyncio
async def test_create_delete_node():
    TEST_NODE_NAME = 'test_node'

    with contextlib.suppress(JenkinsNotFoundError):
        await jenkins.nodes.delete(TEST_NODE_NAME)

    nodes_list = await jenkins.nodes.get_all()
    assert TEST_NODE_NAME not in nodes_list

    config = {
        'name': TEST_NODE_NAME,
        'nodeDescription': '',
        'numExecutors': 10,
        'remoteFS': '',
        'labelString': '',
        'launcher': {
            'stapler-class': 'hudson.slaves.JNLPLauncher',
        },
        'retentionStrategy': {
            'stapler-class': 'hudson.slaves.RetentionStrategy$Always',
        },
        'nodeProperties': {
            'stapler-class-bag': 'true'
        }
    }

    await jenkins.nodes.create(TEST_NODE_NAME, config)
    nodes_list = await jenkins.nodes.get_all()
    assert TEST_NODE_NAME in nodes_list

    with pytest.raises(JenkinsError):
        await jenkins.nodes.create(TEST_NODE_NAME, config)

    await jenkins.nodes.delete(TEST_NODE_NAME)
    nodes_list = await jenkins.nodes.get_all()
    assert TEST_NODE_NAME not in nodes_list



import asyncio
import urllib

from http import HTTPStatus
from typing import Tuple, NamedTuple

import aiohttp

from aiojenkins.exceptions import (
    JenkinsError,
    JenkinsNotFoundError,
)

from aiojenkins.builds import Builds
from aiojenkins.jobs import Jobs
from aiojenkins.nodes import Nodes


class JenkinsVersion(NamedTuple):
    major: int = 0
    minor: int = 0
    patch: int = 0


class Jenkins:

    def __init__(self, host, login=None, password=None):
        self.host = host
        self.auth = None
        self.crumb = None
        self.cookies = None

        if login and password:
            self.auth = aiohttp.BasicAuth(login, password)

        self._nodes = Nodes(self)
        self._jobs = Jobs(self)
        self._builds = Builds(self)

    async def _http_request(self, method: str, path: str, **kwargs):
        if self.auth and not kwargs.get('auth'):
            kwargs['auth'] = self.auth

        if self.crumb:
            kwargs.setdefault('headers', {})
            kwargs['headers'].update(self.crumb)

        try:
            async with aiohttp.ClientSession(cookies=self.cookies) as session:
                response = await session.request(
                    method,
                    urllib.parse.urljoin(self.host, path),
                    allow_redirects=False,
                    **kwargs,
                )
        except aiohttp.ClientError as e:
            raise JenkinsError from e

        if response.status == HTTPStatus.NOT_FOUND:
            raise JenkinsNotFoundError

        if response.status >= HTTPStatus.BAD_REQUEST:
            text = await response.text()

            if response.status in (
                    HTTPStatus.UNAUTHORIZED,
                    HTTPStatus.FORBIDDEN,
                    HTTPStatus.INTERNAL_SERVER_ERROR):
                details = f'probably authentication problem:\n{text}'
            else:
                details = f'\n{text}'

            raise JenkinsError(
                f'Request error [{response.status}], {details}',
                status=response.status,
            )

        if response.cookies:
            self.cookies = response.cookies

        return response

    async def _get_crumb(self):
        try:
            response = await self._http_request('GET', '/crumbIssuer/api/json')
        except JenkinsNotFoundError:
            return False

        data = await response.json()
        self.crumb = {data['crumbRequestField']: data['crumb']}
        return self.crumb

    async def _request(self, method: str, path: str, **kwargs):
        if self.crumb:
            try:
                return await self._http_request(method, path, **kwargs)
            except JenkinsError as e:
                if e.status != 403:
                    raise

        if self.crumb is not False:
            self.crumb = await self._get_crumb()

        return await self._http_request(method, path, **kwargs)

    async def get_status(self) -> dict:
        response = await self._request('GET', '/api/json')
        return await response.json()

    async def get_version(self) -> JenkinsVersion:
        response = await self._request('GET', '/')
        header = response.headers.get('X-Jenkins')
        return JenkinsVersion(*map(int, header.split('.')))

    async def is_ready(self) -> bool:
        """
        Determines is server loaded and ready for work
        """
        try:
            status = await self.get_status()
            return ('mode' in status)
        except JenkinsError:
            return False

    async def wait_until_ready(self, sleep_interval_sec: float = 1.0) -> None:
        """
        Blocks until server is completely loaded
        """
        while (await self.is_ready()) is False:
            await asyncio.sleep(sleep_interval_sec)

    async def quiet_down(self) -> None:
        """
        Start server quiet down period, new builds will not be started
        """
        await self._request('POST', '/quietDown')

    async def cancel_quiet_down(self) -> None:
        """
        Cancel server quiet down period
        """
        await self._request('POST', '/cancelQuietDown')

    async def restart(self) -> None:
        """
        Restart server immediately
        """
        await self._request('POST', '/restart')

    async def safe_restart(self) -> None:
        """
        Restart server when installation is complete and no jobs are running
        """
        await self._request('POST', '/safeRestart')

    @staticmethod
    def _build_token_url(do: str) -> str:
        return f'/me/descriptorByName/jenkins.security.ApiTokenProperty/{do}'

    async def generate_token(self, name: str) -> Tuple[str, str]:
        """
        Generate new API token.

        Returns two values:
        * tokenValue - uses for authorization
        * tokenUuid - uses for revoke
        """
        params = {'newTokenName': name}

        response = await self._request(
            'POST',
            self._build_token_url('generateNewToken'),
            params=params
        )

        response = await response.json()
        if response['status'] != 'ok':
            raise JenkinsError(f'Non OK status returned: `{response}`')

        return response['data']['tokenValue'], response['data']['tokenUuid']

    async def revoke_token(self, token_uuid: str) -> None:
        """
        Revoke API token, please note that uuid is used, not value.
        """
        params = {'tokenUuid': token_uuid}

        await self._request(
            'POST',
            self._build_token_url('revoke'),
            params=params
        )

    @property
    def nodes(self):
        return self._nodes

    @property
    def jobs(self):
        return self._jobs

    @property
    def builds(self):
        return self._builds



import contextlib
import pytest

from aiojenkins.exceptions import (
    JenkinsError,
    JenkinsNotFoundError,
)

from aiojenkins.utils import construct_job_config

from tests import jenkins


TEST_JOB_NAME = 'test'


@pytest.mark.asyncio
async def test_delete_job():
    try:
        await jenkins.jobs.delete(TEST_JOB_NAME)
    except JenkinsNotFoundError:
        ...


@pytest.mark.asyncio
async def test_create_job():
    await jenkins.jobs.create(TEST_JOB_NAME, construct_job_config())


@pytest.mark.asyncio
async def test_get_job_config():
    await jenkins.jobs.get_config(TEST_JOB_NAME)


@pytest.mark.asyncio
async def test_disable_job():
    await jenkins.jobs.disable(TEST_JOB_NAME)


@pytest.mark.asyncio
async def test_disable_unavailable_job():
    with pytest.raises(JenkinsNotFoundError):
        await jenkins.jobs.disable('unavailable')


@pytest.mark.asyncio
async def test_enable_job():
    await jenkins.jobs.enable(TEST_JOB_NAME)


@pytest.mark.asyncio
async def test_get_job_info():
    info = await jenkins.jobs.get_info(TEST_JOB_NAME)
    assert isinstance(info, dict)


@pytest.mark.asyncio
async def test_copy_job():
    job_name_old = test_copy_job.__name__
    job_name_new = job_name_old + '_new'

    with contextlib.suppress(JenkinsError):
        await jenkins.jobs.delete(job_name_old)
        await jenkins.jobs.delete(job_name_new)

    await jenkins.jobs.create(job_name_old, construct_job_config())
    available_jobs = await jenkins.jobs.get_all()
    assert job_name_old in available_jobs

    await jenkins.jobs.copy(job_name_old, job_name_new)
    available_jobs = await jenkins.jobs.get_all()

    assert job_name_new in available_jobs
    assert job_name_old in available_jobs

    await jenkins.jobs.delete(job_name_old)
    await jenkins.jobs.delete(job_name_new)


@pytest.mark.asyncio
async def test_rename_job():
    job_name_old = test_rename_job.__name__
    job_name_new = job_name_old + '_new'

    with contextlib.suppress(JenkinsError):
        await jenkins.jobs.delete(job_name_old)
        await jenkins.jobs.delete(job_name_new)

    await jenkins.jobs.create(job_name_old, jenkins.jobs.construct())
    available_jobs = await jenkins.jobs.get_all()
    assert job_name_old in available_jobs

    await jenkins.jobs.rename(job_name_old, job_name_new)
    available_jobs = await jenkins.jobs.get_all()

    assert job_name_new in available_jobs
    assert job_name_old not in available_jobs

    await jenkins.jobs.delete(job_name_new)


def test_construct_job_config():
    assert len(construct_job_config()) > 0



#!/usr/bin/python3

import asyncio
import contextlib
import pytest

from tests import (
    get_host,
    get_login,
    is_locally,
    jenkins,
)

from aiojenkins import Jenkins
from aiojenkins.exceptions import JenkinsError, JenkinsNotFoundError


@pytest.mark.asyncio
async def test_authentication_error():
    jenkins = Jenkins(get_host(), 'random-login', 'random-password')

    with pytest.raises(JenkinsError):
        version = await jenkins.get_version()
        # was introduced  default admin with password
        if version.major >= 2:
            await jenkins.nodes.disable('master')
        else:
            raise JenkinsError


@pytest.mark.asyncio
async def test_invalid_host():
    with pytest.raises(JenkinsError):
        jenkins = Jenkins('@#$')
        await jenkins.get_version()


@pytest.mark.asyncio
async def test_get_status():
    await jenkins.get_status()


@pytest.mark.asyncio
async def test_quiet_down():
    await jenkins.quiet_down()
    server_status = await jenkins.get_status()
    assert server_status['quietingDown'] is True

    await jenkins.cancel_quiet_down()
    server_status = await jenkins.get_status()
    assert server_status['quietingDown'] is False


@pytest.mark.asyncio
async def test_restart():
    if is_locally():
        pytest.skip('takes too much time +40 seconds')

    await jenkins.safe_restart()
    await asyncio.sleep(5)

    await jenkins.wait_until_ready()
    assert (await jenkins.is_ready()) is True

    await jenkins.restart()
    await jenkins.wait_until_ready()
    assert (await jenkins.is_ready()) is True


@pytest.mark.asyncio
async def test_tokens():
    version = await jenkins.get_version()

    if not (version.major >= 2 and version.minor >= 129):
        pytest.skip('Version isn`t support API tokens')

    token_name = test_tokens.__name__
    job_name = test_tokens.__name__

    token_value, token_uuid = await jenkins.generate_token('')

    token_value, token_uuid = await jenkins.generate_token(token_name)

    await jenkins.nodes.enable('master')

    with contextlib.suppress(JenkinsNotFoundError):
        await jenkins.jobs.delete(job_name)

    await jenkins.jobs.create(job_name, jenkins.jobs.construct())

    # instance without credentials
    jenkins_tokened = Jenkins(get_host(), get_login(), token_value)
    await jenkins_tokened.builds.start(job_name)

    await jenkins.revoke_token(token_uuid)

    with pytest.raises(JenkinsError):
        await jenkins_tokened.builds.start(job_name)

    await jenkins.jobs.delete(job_name)



import json

from aiojenkins.exceptions import (
    JenkinsError,
    JenkinsNotFoundError,
)


class Nodes:

    def __init__(self, jenkins):
        self.jenkins = jenkins

    @staticmethod
    def _normalize_name(name: str) -> str:
        # embedded node `master` actually have brackets in HTTP requests
        if name == 'master':
            return '(master)'
        return name

    async def get_all(self) -> dict:
        """
        Get available nodes on server. Example dict result:
        {
            "master": dict(...),
            "buildbot1": dict(...)
        }
        """
        response = await self.jenkins._request(
            'GET',
            '/computer/api/json'
        )

        nodes = await response.json()
        return {v['displayName']: v for v in nodes['computer']}

    async def get_info(self, name: str) -> dict:
        name = self._normalize_name(name)
        response = await self.jenkins._request(
            'GET',
            f'/computer/{name}/api/json',
        )
        return await response.json()

    async def is_exists(self, name: str) -> bool:
        if name == '':
            return False

        try:
            await self.get_info(name)
        except JenkinsNotFoundError:
            return False
        return True

    async def create(self, name: str, config: dict) -> None:
        if name in await self.get_all():
            raise JenkinsError(f'Node `{name}` is already exists')

        if 'type' not in config:
            config['type'] = 'hudson.slaves.DumbSlave'

        config['name'] = name

        params = {
            'name': name,
            'type': config['type'],
            'json': json.dumps(config)
        }

        await self.jenkins._request(
            'POST',
            '/computer/doCreateItem',
            params=params,
        )

    async def delete(self, name: str) -> None:
        name = self._normalize_name(name)
        await self.jenkins._request(
            'POST',
            f'/computer/{name}/doDelete'
        )

    async def enable(self, name: str) -> None:
        info = await self.get_info(name)
        if not info['offline']:
            return

        name = self._normalize_name(name)
        await self.jenkins._request(
            'POST',
            f'/computer/{name}/toggleOffline'
        )

    async def disable(self, name: str, message: str = '') -> None:
        info = await self.get_info(name)
        if info['offline']:
            return

        name = self._normalize_name(name)
        await self.jenkins._request(
            'POST',
            f'/computer/{name}/toggleOffline',
            params={'offlineMessage': message}
        )

    async def update_offline_reason(self, name: str, message: str) -> None:
        name = self._normalize_name(name)
        await self.jenkins._request(
            'POST',
            f'/computer/{name}/changeOfflineCause',
            params={'offlineMessage': message}
        )



import json

from typing import Any


class Builds:

    def __init__(self, jenkins):
        self.jenkins = jenkins

    async def get_all(self, name: str) -> list:
        """
        Get list of builds for specified job name. Returned example:

        builds = [
          {'number': 1, 'url': 'http://localhost/job/test/1/'},
          {'number': 2, 'url': 'http://localhost/job/test/2/'}
        ]
        """
        response = await self.jenkins._request(
            'GET',
            f'/job/{name}/api/json?tree=allBuilds[number,url]'
        )

        return (await response.json())['allBuilds']

    async def get_info(self, name: str, build_id: int) -> dict:
        """
        Get detailed information about specified build of job
        """
        response = await self.jenkins._request(
            'GET',
            f'/job/{name}/{build_id}/api/json'
        )

        return await response.json()

    async def get_output(self, name: str, build_id: int) -> str:
        """
        Get console output of specified build
        """
        response = await self.jenkins._request(
            'GET',
            f'/job/{name}/{build_id}/consoleText'
        )

        return await response.text()

    async def start(self,
                    name: str,
                    parameters: dict = None,
                    delay: int = 0
                    ) -> None:
        """
        Enqueue new build with delay (default is 0 seconds, means immediately)

        Note about delay (quiet-period):
        https://www.jenkins.io/blog/2010/08/11/quiet-period-feature/
        """
        data = None
        if parameters:
            formatted_parameters: Any = [
                {'name': k, 'value': str(v)} for k, v in parameters.items()
            ]

            if len(formatted_parameters) == 1:
                formatted_parameters = formatted_parameters[0]

            data = {
                'json': json.dumps({
                    'parameter': formatted_parameters,
                    'statusCode': '303',
                    'redirectTo': '.',
                })
            }

        params = {'delay': delay}

        await self.jenkins._request(
            'POST',
            f'/job/{name}/build',
            params=params,
            data=data,
        )

    async def stop(self, name: str, build_id: int) -> None:
        await self.jenkins._request(
            'POST',
            f'/job/{name}/{build_id}/stop'
        )

    async def delete(self, name: str, build_id: int) -> None:
        await self.jenkins._request(
            'POST',
            f'/job/{name}/{build_id}/doDelete'
        )
